/*
 * blueMail offline mail reader
 * some useful auxiliary routines

 Copyright (c) 1996 Toth Istvan <stoty@vma.bme.hu>
 Copyright (c) 1998 William McBrine <wmcbrine@clark.net>
 Copyright (c) 2002 Ingo Brueckl <ib@wupperonline.de>

 Distributed under the GNU General Public License.
 For details, see the file COPYING in the parent directory. */


#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <stdio.h>
#include <time.h>
#include <utime.h>
#include "auxil.h"
#include "../bluemail/bmail.h"


extern bmail bm;


// take off control characters from a string and spaces from its end
char *cropesp (char *st)
{
  char *p = st;

  while (*p)
  {
    if (*p && ((unsigned char) *p < ' ')) *p = ' ';
    p++;
  }
  for (p = strrchr(st, '\0') - 1; p > st && *p == ' '; p--);

  *++p = '\0';
  return st;
}


// duplicate a string
char *strdupplus (const char *original)
{
  char *dup = NULL;

  if (original)
  {
    dup = new char[strlen(original) + 1];
    strcpy(dup, original);
  }

  return dup;
}


// allocate memory for a string
char *stralloc (const char *original)
{
  return (original ? new char[strlen(original) + 1] : NULL);
}


// return a filename without extension
char *stem (const char *fileName)
{
  int e, b;
  static char lcnam[13];

  for (e = 0; fileName[e] && (fileName[e] != '.'); e++);
  for (b = 0; b < e; b++) lcnam[b] = fileName[b];
  lcnam[b] = '\0';

  return lcnam;
}


// get rid of RE:
const char *stripRE (const char *subject)
{
  while (strncasecmp(subject, "RE: ", 4) == 0) subject += 4;
  return subject;
}


// delete any file in a directory
void clearDirectory (const char *dirname)
{
  DIR *dir;
  struct dirent *entry;

  if ((dir = opendir(dirname)))
  {
    if (mychdir(dirname) == 0)
      while ((entry = readdir(dir)))
        if (*entry->d_name != '.') remove(entry->d_name);
    closedir(dir);
  }
}


// to avoid segmentation faults
size_t slen (const char *s)
{
  return (s ? strlen(s) : 0);
}


// build a path to a filename in a directory
int mkfname (char *dest, const char *dir, const char *file)
{
  char format[12];

#ifdef __MSDOS__
  #define FORMAT  "%%.%lus/%%s"
#else
  #define FORMAT  "%%.%us/%%s"
#endif

  sprintf(format, FORMAT, MYMAXPATH - strlen(file) - 2);
  return sprintf(dest, format, dir, file);
}


// build a filename from stem and extension
int mkfnamext (char *dest, const char *fstem, const char *fext)
{
  int result = mkfname(dest, fstem, fext);
  char *delim = strrchr(dest, '/');
  *delim = '.';
  return result;
}


#ifdef DOSPATHTYPE
// convert pathnames to "canonical" form (exchange slashes and backslashes)
char *convert (const char *sinner, char from, char to)
{
  static char saint[MYMAXPATH];
  int i;

  for (i = 0; sinner[i]; i++) saint[i] = (sinner[i] == from ? to : sinner[i]);

  saint[i] = '\0';
  return saint;
}
#endif


/* Copy a character to position pos into buffer, but not more than maxchar
   to prevent buffer from overflowing. Buffer must have length maxchar+1.
   If ok, return next pos. */
char *charcpy (const char *buffer, char *pos, unsigned int maxchar, char c)
{
  *pos = '\0';
  if (strlen(buffer) < maxchar) *pos++ = c;
  return pos;
}


// smart strtok routine
char *strtoken (char *string, char delim, bool left)
{
  static char buffer[MYMAXLINE];
  char *str, *buf;

  // lame, but protects against buffer overflow
  if (strlen(string) >= sizeof(buffer))
    return (left ? string : &string[strlen(string)]);

  str = string;
  buf = buffer;

  while (*str && (*str != delim)) *buf++ = *str++;
  *buf = '\0';
  if (*str) str++;

  if (left) return buffer;
  else return str;
}


// basically the same as strtok, but returns empty tokens, too
char *strtokn (char *string, char delim)
{
  static char *last, *next;

  if (string) last = next = string;     // initialize
  else last = next;

  while (*next && (*next != delim)) next++;

  if (*next) *next++ = '\0';

  return last;
}


// build a command string with argument
char *makecmd (const char *cmd, const char *arg)
{
  static char command[MYMAXPATH + MYMAXPATH];
  char format[18];

  sprintf(format, "%%.%us %%.%us", MYMAXPATH - 1, MYMAXPATH - 1);
  sprintf(command, format, cmd, arg);
  return command;
}


// expand ~/ in pathnames
const char *tilde_expand (const char *path)
{
  const char *home = mygetenv ("HOME");
  static char result[MYMAXPATH];

  if (path && home && (*path == '~') && (path[1] == '/'
#ifdef DOSPATHTYPE
                                                    || path[1] == '\\'
#endif
                                                                      ) &&
      (strlen(home) + strlen(path) - 1 < MYMAXPATH))
  {
    sprintf(result, "%s%s", home, path + 1);
    return result;
  }
  else return path;
}


// for consistency, no path should end in a slash
const char *fixPath (const char *path)
{
  size_t l;
  static char result[MYMAXPATH];

  if ((l = strlen(path)) > 0)
  {
    char d = path[l - 1];

    if (d == '/'
#ifdef DOSPATHTYPE
                 || d == '\\'
#endif
                             )
    {
      strncpy(result, path, MYMAXPATH);
      result[(l < MYMAXPATH ? l : MYMAXPATH) - 1] = '\0';
      return result;
    }
  }

  return path;
}


// append '.' if file has no extension
const char *extent (const char *path)
{
  bool charfound = false;
  size_t len = strlen(path);
  static char result[MYMAXPATH];

  for (int i = len - 1; i >= 0; i--)
  {
    if (path[i] == '/'
#ifdef DOSPATHTYPE
                       || path[i] == '\\'
#endif
                                         ) break;
    if (path[i] == '.') return path;
    charfound = true;
  }

  strcpy(result, path);
  if (charfound && (len < MYMAXPATH - 1)) strcat(result, ".");
  return result;
}


// get a little endian short as integer
unsigned int getLSBshort (const unsigned char *x)
{
  return ((unsigned int) x[1] << 8) + (unsigned int) x[0];
}


// get a little endian long as long
unsigned long getLSBlong (const unsigned char *x)
{
  return ((unsigned long) x[3] << 24) + ((unsigned long) x[2] << 16) +
         ((unsigned long) x[1] << 8) + (unsigned long) x[0];
}


// get a big endian long as long
unsigned long getMSBlong (const unsigned char *x)
{
  return ((unsigned long) x[0] << 24) + ((unsigned long) x[1] << 16) +
         ((unsigned long) x[2] << 8) + (unsigned long) x[3];
}


// put an integer into a little endian short
void putLSBshort (unsigned char *dest, unsigned int source)
{
  dest[0] = source & 0xFF;
  dest[1] = (source & 0xFF00) >> 8;
}


// put a long into a little endian long
void putLSBlong (unsigned char *dest, unsigned long source)
{
  dest[0] = source & 0xFF;
  dest[1] = (source & 0xFF00) >> 8;
  dest[2] = (source & 0xFF0000) >> 16;
  dest[3] = (source & 0xFF000000) >> 24;
}


// put a long into a big endian long
void putMSBlong (unsigned char *dest, unsigned long source)
{
  dest[0] = (source & 0xFF000000) >> 24;
  dest[1] = (source & 0xFF0000) >> 16;
  dest[2] = (source & 0xFF00) >> 8;
  dest[3] = source & 0xFF;
}


// put a float in Microsoft MKS binary format into a long
unsigned long getMKSlong (const unsigned char *mks)
{
  unsigned long m = (unsigned long) mks[0] +
                    ((unsigned long) mks[1] << 8) +
                    ((unsigned long) mks[2] << 16);

  return (m | 0x800000L) >> (24 + 0x80 - mks[3]);
}


// return file modification time
time_t fmtime (const char *file)
{
  struct stat fstat;

  return (stat(file, &fstat) == 0 ? fstat.st_mtime : 0);
}


// set a file's modification time
void fmtime (const char *file, time_t mtime)
{
  struct utimbuf ut;

  time(&ut.actime);
  ut.modtime = mtime;
  utime(file, &ut);
}


// return file size
off_t fsize (const char *file)
{
  struct stat fstat;

  return (stat(file, &fstat) == 0 ? fstat.st_size : -1);
}


// return the filename or directory part of a path
const char *fsplit (const char *path, bool filename)
{
  static char dir[MYMAXPATH + 1];
  strncpy (dir, path, MYMAXPATH);
  dir[MYMAXPATH] = '\0';
  int slen = strlen(dir);

  while (--slen >= 0)
  {
    char p = dir[slen];
    if (p == '/' || p == '\\') break;
  }

  if (filename) return dir + slen + 1;
  else
  {
    if (slen < 0) dir[++slen] = '.';
    if (slen == 0) slen++;
    dir[slen] = '\0';
    return dir;
  }
}


// create an empty file in a directory
void fcreate (const char *dir, const char *file)
{
  char fname[MYMAXPATH];
  FILE *f;

  mkfname(fname, dir, file);
  if ((f = fopen(fname, "wb"))) fclose(f);
}


// copy a file, return error(level)
int fcopy (const char *from, const char *to)
{
  FILE *src = NULL, *dest = NULL;
  size_t len;

  src = fopen(from, "rb");
  dest = fopen(to, "wb");

  if (!src || !dest)
  {
    if (src) fclose(src);
    if (dest) fclose(dest);
    return 1;   // error(level)
  }

  char *buffer = new char[BLOCKLEN];

  while ((len = fread(buffer, sizeof(char), BLOCKLEN, src)))
    fwrite(buffer, sizeof(char), len, dest);

  delete[] buffer;
  fclose(src);
  fclose(dest);

  return 0;   // error(level)
}


// get rid of newline (and cr) character
char *mkstr (char *line)
{
  char *pos = strrchr(line, '\n');

  if (pos)
  {
    *pos-- = '\0';
    if (pos >= line && *pos == '\r') *pos = '\0';
  }

  return line;
}


// convert a string to upper case
char *strupper (char *s)
{
  char *p = s;

  while (*p)
  {
    *p = toupper(*p);
    p++;
  }
  return s;
}


// change fname's extension to new_ext
char *ext (const char *fname, const char *new_ext)
{
  static char result[13];

  sprintf(result, "%.8s.%.3s", stem(fname), new_ext);
  return result;
}


// check whether filename contains wildcard and tell fname's length
int wildcard (const char *fname)
{
  int pos = strlen(fname) - 1;

  return (pos >= 0 && fname[pos] == '*' ? pos : -1);
}


// MIME charsets we can handle
CHARSETMAP charsetmap[] = {{"ISO-8859-1", true}, {"ISO-8859-15", true},
                           {"WINDOWS-1252", true}, {"US-ASCII", false},
                           {"IBM437", false}, {"CP437", false},
                           {"X-CP437", false}, {NULL, false}};

// decode a RFC2047 encoded header field
// and report whether encoding was ISO 8859-1
void mimeDecodeHeader (const char *source, char *dest, bool *latin1)
{
  int i, qp, b64idx = 0;
  char c, t = '\0', b[4];
  const char *cset, *b64char;
  size_t lenc, lens;
  char *endpos = NULL;
  bool insideQ = false, insideB = false;

  // Base64 alphabet
  static const char *b64alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                                   "abcdefghijklmnopqrstuvwxyz"
                                   "0123456789+/";

  if (source)
  {
    while (*source)
    {
      c = *source++;

      // start of encoding?
      if ((c == '=' && *source == '?') && !insideB)
      {
        i = 0;
        lens = strlen(source);

        while ((cset = charsetmap[i++].charset))
        {
          lenc = strlen(cset);

          if (// source long enough
              lens >= lenc + 4 &&
              // known character set
              strncasecmp(source + 1, cset, lenc) == 0 &&
              // well-formed
              source[lenc + 1] == '?' && source[lenc + 3] == '?' &&
              // encoding type Q or B
              ((t = toupper(source[lenc + 2])) == 'Q' || t == 'B'))
          {
             if (t == 'Q') insideQ = true;
             else insideB = true;

             if (endpos) dest = endpos;
             endpos = NULL;

             b64idx = 0;
             source += lenc + 4;
             *latin1 = charsetmap[--i].isLatin1;
             break;
          }
        }

        // wasn't start of encoding
        if (!cset) *dest++ = c;
      }

      // end of encoding?
      else if (c == '?' && *source == '=')
      {
        if (insideQ || insideB)
        {
          source++;
          endpos = dest;
        }
        else
        {
          *dest++ = c;
          endpos = NULL;
        }

        insideQ = insideB = false;
      }

      // inside quoted-printable?
      else if (insideQ)
      {
        switch (c)
        {
          case '=':
            qp = 0;
            b[1] = b[2] = '\0';
            b[0] = *source;
            if (*source) b[1] = *++source;
            sscanf(b, "%x", (unsigned int *) &qp);
            if (qp != 0) *dest++ = qp;
            if (*source) source++;
            break;

          case '_':
            *dest++ = ' ';
            break;

          default:
            *dest++ = c;
        }
      }

      // inside base64?
      else if (insideB)
      {
        if ((b64char = strchr(b64alphabet, c)))
        {
          b[b64idx] = b64char - b64alphabet;

          switch (b64idx++)
          {
            case 0:
              break;

            case 1:
              *dest++ = (b[0] << 2 | b[1] >> 4);
              break;

            case 2:
              *dest++ = (b[1] << 4 | b[2] >> 2);
              break;

            case 3:
              *dest++ = (b[2] << 6 | b[3]);
              b64idx = 0;
              break;
          }
        }
      }

      // not an encoded part
      else
      {
        *dest++ = c;
        if (c != ' ' && c != '\t') endpos = NULL;
      }
    }
  }

  *dest = '\0';
}


// get rightmost word in string
const char *strrword (const char *str)
{
  const char *p = strrchr(str, ' ');
  return (p ? ++p : str);
}


// upper and lower case conversion for non-ASCII characters
char uplow (char c, int look, int take)
{
  const char *map = bm.resourceObject->get(UpperLower);
  // this map must have an even length

  if (*map)
    for (int i = 0; map[i]; i += 2)
      if (map[i + look] == c) return map[i + take];

  return '\0';
}


// strncasecmp() (taken from glibc 2.2.5) for non-ASCII characters
int strniccmp (const char *s1, const char *s2, size_t n)
{
  const char *map = bm.resourceObject->get(UpperLower);

  if (*map)
  {
    const unsigned char *p1 = (const unsigned char *) s1;
    const unsigned char *p2 = (const unsigned char *) s2;
    unsigned char _c, c1, c2;

    if (p1 == p2 || n == 0) return 0;

    do
    {
      c1 = to_lower(*p1);
      p1++;
      c2 = to_lower(*p2);
      p2++;

      if (c1 == '\0' || c1 != c2) return c1 - c2;
    }
    while (--n > 0);

    return c1 - c2;
  }
  else return strncasecmp(s1, s2, n);
}


// case-insensitive version of strstr()
char *stricstr (const char *s1, const char *s2)
{
  bool found = false;
  char _c, start[3];
  char *str = NULL;

  if (*s2)
  {
    size_t len = strlen(s2) - 1;

    start[0] = to_upper(*s2);
    start[1] = to_lower(*s2);
    start[2] = '\0';

    s2++;

    do
    {
      if ((str = strpbrk(s1, start)))
      {
        s1 = str + 1;
        found = (strniccmp(s1, s2, len) == 0);
      }
    }
    while (str && !found && *s1);
  }

  return (found ? str : NULL);
}


// check whether two hex digits follow
bool ishex (char *digits)
{
  if (digits[0] == '\0' || digits[1] == '\0') return false;
  else return (isxdigit((unsigned char) digits[0]) &&
               isxdigit((unsigned char) digits[1]));
}


// basically the same as fgets, but if the buffer is too small (i.e. '\n'
// was not read from stream, it reads on until the end of line is reached;
// returns count of characters read
size_t fgetsnl (char *s, int n, FILE *stream)
{
  int c;

  size_t len = slen(fgets(s, n, stream));

  if (len && !strchr(s, '\n'))
    do
    {
      if ((c = fgetc(stream)) == EOF) break;
      else len++;
    }
    while (c != '\n');

  return len;
}


// parse an option line, return name and value
void parseOption (char *line, char delim, char *&name, char *&value)
{
  char *pos = line;

  // leading spaces
  while (*pos && (*pos == ' ' || *pos == '\t')) pos++;

  // option name
  name = pos;
  while (*pos && (*pos != delim && *pos != ' ' && *pos != '\t')) pos++;

  if (*pos) *pos++ = '\0';

  // chars between strings
  while (*pos && (*pos == ' ' || *pos == '\t' || *pos == delim)) pos++;

  // option value
  value = pos;
  cropesp(value);
}


// find name part and address part of address
void parseAddress (const char *na, int &name, int &n_len,
                                   int &addr, int &a_len)
{
  char *pos;

  n_len = a_len = 0;

  if ((pos = strchr(na, '(')))
  {
    name = pos - na + 1;
    addr = 0;
    a_len = pos - na - 1;
    while (*pos && (*pos != ')'))
    {
      n_len++;
      pos++;
    }
    n_len--;
  }
  else if ((pos = strchr(na, '<')))
  {
    name = 0;
    addr = pos - na + 1;
    n_len = addr - 2;
    while (*pos && (*pos != '>'))
    {
      a_len++;
      pos++;
    }
    a_len--;

    if (*na == '"')
    {
      name++;
      n_len--;
    }
    int end = n_len - 1;
    if (end >= 0 && na[name + end] == '"') n_len--;
  }
  else if ((pos = strchr(na, '@')))
  {
    name = addr = 0;
    a_len = strlen(na);
  }
  else
  {
    name = addr = 0;
    n_len = strlen(na);
  }

  if (n_len < 0) n_len = 0;
  if (a_len < 0) a_len = 0;
}


// simple check
bool isEmailAddress (const char *addr, int len)
{
  const char *at = NULL, *dot = NULL;

  if (len > 0)
    for (int i = 0; i < len; i++)
    {
      if (addr[i] == '@')
      {
        if (!at) at = addr + i;
        else
        {
          at = NULL;
          break;
        }
      }
      if (addr[i] == '.') dot = addr + i;
    }

  return (at && dot && (at > addr) && (dot > at + 1) && (dot < addr + len - 1));
}
